/*
 * Log.java
 * by Frederik Schorr
 * with changes by Geometrek
 *  This file is part of DeepMatrixIP9.
 *
 * 
 * Copyright (C) 1999 Geometrek	and 2009 by R. Russell Kinter
 *
 *  This file is part of DeepMatrixIP9.
 *
 *  DeepMatrixIP9 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  DeepMatrix IP9 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package matrix.server;

import matrix.*;
import matrix.vrml.*;
import java.io.*;
import java.text.*;
import java.util.*;

/**
 * Logs messages with timestamps and thread ids to file.
 * This class was developed by Frederik Schorr during a CS course.
 * I added some minor changes.
 */
public class ObjectRecord implements Runnable {
    private boolean kill = false;
    private boolean exists = false;
    private boolean change = false;
    private Room RM;
    private File OrigFile;
    private File TempFile;
    private Hashtable ObjStore = new Hashtable();
    private String RoomName;
    private int Roomnum = 0;
    private String objectrecord;
    private Vector buffer = new Vector();
    private Vector OIvec  = new Vector();
    //private String target;
    private Thread runner; 
    private BufferedReader Breader = null;
	private PrintWriter Pwriter = null;
	
	final int BUFFERSIZE = 4096;
	

	public ObjectRecord(String dir, String RN, Room rm, int num ) {
         RM = rm;
         Roomnum = num;
		 objectrecord = ( dir +  RN + File.separatorChar + RN + "_Objects.txt" );
		 TempFile = new File(dir +  RN + File.separatorChar + RN + "_TempFile.txt");
         RoomName = RN;
	     runner = new Thread(this,RoomName + "_OR_" + Roomnum);
         start();
	}

    public void start(){
    if(runner!=null){ runner.start(); }
    }

    public void run(){
			 ControlMessage cm;
              ObjEnvelope oe;
			  Message msg;
			readInfo( objectrecord );
            while( Thread.currentThread() == runner ){
                     try{
                        synchronized( buffer ){
						    if(!buffer.isEmpty()){
                           String sObj;
                           String sSender;
                             oe = (ObjEnvelope) buffer.firstElement();
                            msg = oe.RTS;
						  	switch( msg.getType()) {
			             	case RealTimeSave.TYPE :
                          RealTimeSave rts = (RealTimeSave)msg; 
                          sObj = rts.tag;
                          sSender = oe.sender;
                    		 if(ObjStore.containsKey( rts.tag )){
						   float [] mft = new float[7];
                           int iState = ((SFInt32)rts.state).getValue();
	                       ObjInfo OI2 =  (ObjInfo)ObjStore.get( sObj );
                         if(OI2.objState==0){
				         OI2.owner = sSender;
				         mft = ((MFFloat)rts.data).getValue();
    					 OI2.position[0] = mft[0];
                         OI2.position[1] = mft[1];
						 OI2.position[2] = mft[2];
                         OI2.rotation[0] = mft[3];
                		 OI2.rotation[1] = mft[4];
                         OI2.rotation[2] = mft[5];
                         OI2.rotation[3] = mft[6];
              cm = new ControlMessage( ControlMessage.OBJREQ, "MatrixD", "Object " + OI2.objName + " will be loaded into this room!");
		 RM.postTo( sSender, cm );
		 cm	 = new ControlMessage( ControlMessage.OBJADD, OI2.objName, OI2.objUrl );
       System.out.println("OR requestObject: " + OI2.objName );
    	 RM.postTo( sSender, cm );
             OI2.objState = iState;
           }
        else {
        if(OI2.owner.equals( sSender )||OI2.owner.equals("")){
          if(iState==0){ 
            OI2.resetFloats();
		 cm = new ControlMessage( ControlMessage.OBJREQ, "MatrixD", "Object " + OI2.objName + " will be unloaded from this room!");
		 RM.postTo( sSender, cm );
         cm	 = new ControlMessage( ControlMessage.OBJDEL, OI2.objName, OI2.objUrl ); 
    	 RM.postToAll( cm );
          OI2.objState = iState;
           }
           else {
           mft = ((MFFloat)rts.data).getValue();
    	OI2.position[0] = mft[0];
                         OI2.position[1] = mft[1];
						 OI2.position[2] = mft[2];
                         OI2.rotation[0] = mft[3];
                		 OI2.rotation[1] = mft[4];
                         OI2.rotation[2] = mft[5];
                         OI2.rotation[3] = mft[6];
		   RealTimeMessage rtm = new RealTimeMessage( sObj, new SFVec3f(mft[0],mft[1],mft[2]), true );
             RM.postToAll( rtm );
	RealTimeMessage rtm2 = new RealTimeMessage( sObj, new SFRotation(mft[3],mft[4],mft[5],mft[6]), true );
             RM.postToAll( rtm2 );
 cm = new ControlMessage( ControlMessage.MSG, "MatrixD", "Info for Object " + sObj + " saved and updated!");
             RM.postTo( sSender, cm );
                }

        }
     else {
 cm = new ControlMessage( ControlMessage.MSG, "MatrixD", "Object " + sObj + " is currently controlled by " + OI2.owner + "!");
             RM.postTo( sSender, cm );
     }
   }
  }
  else {
	cm = new ControlMessage( ControlMessage.ERROR, "MatrixD", "Can't save info for Object " + sObj + " does not exist in this room!");
             RM.postTo( sSender, cm );
     }
    break;
  	default :
	 RM.exceptionLog("ObjectRecord Object: unknown message type " + msg.getType() );
    break;
      } 
     buffer.removeElementAt(0);
       } 
    else {
	Enumeration list2 = ObjStore.elements();
        	  while( list2.hasMoreElements()){
       ObjInfo OI3 = ((ObjInfo)list2.nextElement());
       System.out.println(OI3.printObjInfo());
       }
   System.out.println("OR " + RoomName + " is going to wait");
                 if(kill==false){
                      buffer.wait();
                           }
                             else {
			           if(Pwriter!=null){  
                         Pwriter.flush();
			             Pwriter.close();
                         }
                         runner = null;
                            }
                          }
                        }
                    }
                catch(InterruptedException e){}
                catch(NullPointerException f){
                RM.exceptionLog(" " + RoomName + "_OR_" + Roomnum + " : " + f + " "); }
                catch(Exception g){
                RM.exceptionLog(" " + RoomName + "_OR_" + Roomnum + " : " + g + " ");  
                }
            }
        }

       public void closeWrite(){
		if(exists){
		Enumeration list = ObjStore.elements();
        	  while( list.hasMoreElements()){
       ObjInfo OI = ((ObjInfo)list.nextElement());
       String sOi = OI.printObjInfo();
       OI = null;
			try {
		    Pwriter.write( sOi );
      System.out.println(sOi);
			Pwriter.println();	    
			Pwriter.flush();	    
		}
		catch(Exception e) { 
          }
         }
           }
          }


       private void readInfo(String file) {
        int Ct1 = 0;
        int Ct2 = 0;
		int index = 0;
         String line, objName, keyword, objChange, objUrl;
        OrigFile = new File(file);
		exists = OrigFile.exists();
        if(exists){
        try {
			Breader = new BufferedReader( new FileReader(file) );
			line = Breader.readLine();
            while(line != null ) {
                if( line.length() > 0 ) {
                 Ct1 = Ct2;
                 Ct2 += line.length();
                    StringTokenizer token = new StringTokenizer(line," ");
                    keyword = token.nextToken();
                    if(keyword.charAt(0) != '#') {                          // No Comment !
                      String  obj = keyword.toLowerCase();
                        if(obj.equals("object")) { 
                            try { 
                            objName = token.nextToken();
							 }
                            catch(NoSuchElementException e) {
                    RM.exceptionLog(" Error parsing "+file +" : "+line+"\n\t"+e +" ");
                            return;
                            } 
						if( ObjStore.containsKey(objName))
                                    System.err.println(RoomName+" is allready defined!");
                          else {
                            try {
                            objUrl = token.nextToken();
                            
                            }
							catch(NoSuchElementException e) {
                     RM.exceptionLog(" Error parsing "+file +" position: "+line+"\n\t"+e + " ");
                            return;
                            }
                         boolean  bBool = false;
                            try {
                          Boolean BB = new Boolean( token.nextToken() );
						 bBool =  BB.booleanValue();
                             }
						 catch(NoSuchElementException e) {
                     RM.exceptionLog(" Error parsing "+file +" position: "+line+"\n\t"+e + " ");
                            return;
                            }
                            int num = 0;
							try {
							 Integer Int = new Integer( token.nextToken() );
							num = Int.intValue();
                            }
							catch(NoSuchElementException e) {
                     RM.exceptionLog(" Error parsing "+file +" position: "+line+"\n\t"+e + " ");
						    return;
							}
						   	catch(NumberFormatException f){
                      RM.exceptionLog(" Error parsing "+file+" position : "+line+"\n\t"+f+ " ");
                             return;
                              }
							float [] pos = new float [3];
		                    float [] rot = new float [4];
                            if(num!=0){
							for( int i = 0; i<3; i++){
                            try {
                            Float ft = new Float( token.nextToken() );
            				pos[i] = ft.floatValue();
                            }
							catch(NoSuchElementException e) {
                     RM.exceptionLog(" Error parsing "+file +" position: "+line+"\n\t"+e + " ");
	                        return;
                            } 
							catch(NumberFormatException f){
                     RM.exceptionLog(" Error parsing "+file+" position : "+line+"\n\t"+f+ " ");
                             return;
                              }
                            }
                            for( int i = 0; i<4; i++){
                            try {
							Float ft = new Float( token.nextToken() );
            				rot[i] = ft.floatValue();
                             }
							catch(NoSuchElementException e) {
                      String er = " Error parsing "+file +" rotation: "+line+"\n\t"+e + " ";
                      RM.exceptionLog( er );
						     return;
                            } 
							catch(NumberFormatException f){
                      String er = " Error parsing "+file+" rotation : "+line+"\n\t"+f+ " ";
                      RM.exceptionLog( er );
                              return;
                               }
                              }
                            }
                            else {  
						    float [] dP = new float [] { (float)0.0, (float)0.0, (float)0.0 };
                            float [] dR = new float [] { (float)0.0, (float)1.0, (float)0.0, (float)0.0 };
                             pos = dP; rot = dR; }
                         ObjInfo OI = new ObjInfo(objName, objUrl, bBool, num, Ct1, Ct2, pos, rot);
                         ObjStore.put( objName, OI );
                      System.out.println("Done reading line: " +line);
                        }
                       }
                      }
                     }
                line = Breader.readLine();
                index++;
           }
	   	    FileOutputStream fos = new FileOutputStream(OrigFile);
            Pwriter = new PrintWriter(fos);
         }
        catch(Exception e) { 
            String er = " Error reading "+file+" in line "+index + " "; 
            RM.exceptionLog( er ); 
        }
      }
    else {
    String er = " " + file + " does not exist! ";
	RM.exceptionLog( er);
     }
    }

    public String [] getObjis(){
	  	String[] val = new String[(ObjStore.size())];
		int i = 0;
		Enumeration list = ObjStore.elements();
		while( list.hasMoreElements()) {
         ObjInfo OI =  (ObjInfo)list.nextElement();
         if(0<OI.objState){
			val[i] = OI.objName + "(Unload)";
             }
          else{
            val[i] = OI.objName + "(Load)";
           }
          i += 1;
        }
		return val;
   }
   
   public boolean sharedObjectExist( String name ){
	 if( !ObjStore.containsKey( name )){
    System.out.println("OR: sharedObjectExist false " + name);
	  return false;
        }
    else
    System.out.println("OR: sharedObjectExist true " + name);
    return true;
   }
   
   public void changeOwnerObjects( String quiter ){
	   	Enumeration list = ObjStore.elements();
        	  while( list.hasMoreElements()){
       ObjInfo OI3 = ((ObjInfo)list.nextElement());
	  if(OI3.owner.equals(quiter)&&(!OI3.objChange)){
         OI3.owner = "";
        }
      }
   }

   public void addLoadedObjects( String requester ){
	   	Enumeration list = ObjStore.elements();
        	  while( list.hasMoreElements()){
       ObjInfo OI3 = ((ObjInfo)list.nextElement());
	  if(0<OI3.objState){
System.out.println("OR addLoadedObjects: " + OI3.objName + " " + OI3.objUrl);
 ControlMessage	cm	 = new ControlMessage( ControlMessage.OBJADD, OI3.objName, OI3.objUrl );
	 RM.postTo( requester, cm );
        }
      }
   }

   public synchronized void updateSharedObject( String name, String target ){
	if(ObjStore.containsKey( name )){
	ObjInfo OI =  (ObjInfo)ObjStore.get( name );
    if(0<OI.objState){
     Field rot = new SFRotation( OI.rotation );
    RealTimeMessage rtm1 = new RealTimeMessage( name, rot,   false    );
    System.out.println("updateSharedObject SFRotation: " + OI.rotation[0] + " " + OI.rotation[1] + " " + OI.rotation[2] + " " + OI.rotation[3]);
	RM.postToRTM( target, rtm1 );
     Field pos = new SFVec3f( OI.position );
	RealTimeMessage rtm2 = new RealTimeMessage( name, pos, false  );
	 System.out.println("updateSharedObject SFVec3f: " + OI.position[0] + " " +  OI.position[1] + " " +  OI.position[2] );
	RM.postToRTM( target, rtm2 );
    }
    else {
    ControlMessage cm = new ControlMessage( ControlMessage.ERROR, "MatrixD", "Object " + name + " has not yet been loaded!");
    RM.postTo( target, cm );
    }
    }
	else {
    ControlMessage cm = new ControlMessage( ControlMessage.ERROR, "MatrixD", "Object " + name + " does not exist in this room!");
    RM.postTo( target, cm );
    }
   }


	public  void saveObjectInfo( String sender, Message rts ){
   ObjEnvelope OE = new ObjEnvelope( sender, rts );
	   synchronized( buffer ){
                buffer.addElement( OE );
                buffer.notify();
            }
    }

	
	/**
	 * Close log file
	 */
	public void close() {
	   synchronized( buffer ){
                kill = true;
                buffer.removeAllElements();
                buffer.notify();
            }
	
	}
	
	
}
